@@ -65,7 +65,7 @@ pub trait LinkedList: 'static + Sized {
             .update_tag_if(Tag::First, |t| t == Tag::None, order)
     }
 
-    /// Removes marks from `self`.
+    /// Removes the mark from `self`.
     ///
     /// It returns `false` if no flag has been marked on `self`.
     ///
@@ -229,10 +229,13 @@ pub trait LinkedList: 'static + Sized {
             entry
                 .link_ref()
                 .swap((next_ptr.get_arc(), Tag::None), Relaxed);
-            match self
-                .link_ref()
-                .compare_exchange(next_ptr, (Some(entry), new_tag), order, Relaxed)
-            {
+            match self.link_ref().compare_exchange(
+                next_ptr,
+                (Some(entry), new_tag),
+                order,
+                Relaxed,
+                barrier,
+            ) {
                 Ok((_, updated)) => {
                     return Ok(updated);
                 }
@@ -303,6 +306,7 @@ pub trait LinkedList: 'static + Sized {
                 (next_valid_ptr.get_arc(), self_tag),
                 Release,
                 Relaxed,
+                barrier,
             ) {
                 barrier.reclaim(prev);
             }
