@@ -3,6 +3,7 @@ use super::leaf_node::{LOCKED, RETIRED};
 use super::node::{Node, Type};
 
 use crate::ebr::{Arc, AtomicArc, Barrier, Ptr, Tag};
+use crate::wait_queue::WaitQueue;
 
 use std::borrow::Borrow;
 use std::cmp::Ordering::{Equal, Greater, Less};
@@ -29,6 +30,9 @@ where
     /// `latch` acts as a mutex of the [`InternalNode`] that also stores the information about an
     /// on-going structural change.
     latch: AtomicArc<StructuralChange<K, V>>,
+
+    /// `wait_queue` for `latch`.
+    wait_queue: WaitQueue,
 }
 
 impl<K, V> InternalNode<K, V>
@@ -42,6 +46,7 @@ where
             children: Leaf::new(),
             unbounded_child: AtomicArc::null(),
             latch: AtomicArc::null(),
+            wait_queue: WaitQueue::default(),
         }
     }
 
@@ -337,16 +342,20 @@ where
             ),
             Acquire,
             Relaxed,
+            barrier,
         ) {
             debug_assert!(!self.retired(Relaxed));
             if full_node_ptr != full_node.load(Relaxed, barrier) {
-                drop(self.latch.swap((None, Tag::None), Relaxed));
+                let (change, _) = self.latch.swap((None, Tag::None), Relaxed);
+                self.wait_queue.signal();
+                drop(change);
                 target.rollback(barrier);
                 return Err((key, value));
             }
             unsafe { &mut *(ptr.as_raw() as *mut StructuralChange<K, V>) }
         } else {
             target.rollback(barrier);
+            self.wait(barrier);
             return Err((key, value));
         };
 
@@ -537,30 +546,29 @@ where
             }
         };
 
-        // Replaces the full node with the high-key node.
-        let unused_node = full_node.swap(
-            (new_nodes.high_key_node.get_arc(Relaxed, barrier), Tag::None),
-            Release,
-        );
+        // Replace the full node with the high-key node.
+        let unused_node = full_node
+            .swap(
+                (new_nodes.high_key_node.get_arc(Relaxed, barrier), Tag::None),
+                Release,
+            )
+            .0;
 
         if root_split {
             // Return without unlocking it.
             return Err((key, value));
         }
 
-        // Drops the deprecated nodes.
-        // - Still, the deprecated full leaf can be reachable by Scanners.
+        // Unlock the node.
+        self.finish_split(barrier);
+
+        // Drop the deprecated nodes.
         if let Some(unused_node) = unused_node {
-            // Cleans up the split operation by committing it.
+            // Clean up the split operation by committing it.
             unused_node.commit(barrier);
             barrier.reclaim(unused_node);
         }
 
-        // Unlock the node.
-        if let Some(change) = self.latch.swap((None, Tag::None), Release) {
-            barrier.reclaim(change);
-        }
-
         // Traverse several leaf nodes in order to cleanup deprecated links.
         self.max_le_appr(key.borrow(), barrier);
 
@@ -568,17 +576,21 @@ where
         Err((key, value))
     }
 
-    /// Finishes splitting the root node.
-    pub fn finish_root_split(&self, barrier: &Barrier) {
-        if let Some(change) = self.latch.swap((None, Tag::None), Release) {
+    /// Finishes splitting the [`InternalNode`].
+    pub fn finish_split(&self, barrier: &Barrier) {
+        let (change, _) = self.latch.swap((None, Tag::None), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
             barrier.reclaim(change);
         }
     }
 
-    /// Commits an on-going structural change.
+    /// Commits an on-going structural change recursively.
     pub fn commit(&self, barrier: &Barrier) {
-        // Keeps the internal node locked to prevent further locking attempts.
-        if let Some(change) = self.latch.swap((None, LOCKED), Relaxed) {
+        // Mark the internal node retired to prevent further locking attempts.
+        let (change, _) = self.latch.swap((None, RETIRED), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
             let obsolete_node_ptr = change.origin_node.load(Relaxed, barrier);
             if let Some(obsolete_node_ref) = obsolete_node_ptr.as_ref() {
                 obsolete_node_ref.commit(barrier);
@@ -588,25 +600,35 @@ where
 
     /// Rolls back the ongoing split operation recursively.
     pub fn rollback(&self, barrier: &Barrier) {
-        if let Some(change) = self.latch.load(Relaxed, barrier).as_ref() {
-            if let Some(origin) = change.origin_node.swap((None, Tag::None), Relaxed) {
+        let (change, _) = self.latch.swap((None, Tag::None), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
+            if let Some(origin) = change.origin_node.swap((None, Tag::None), Relaxed).0 {
                 origin.rollback(barrier);
             }
-        }
-
-        // Unlocks the node after the origin node has been cleaned up.
-        if let Some(change) = self.latch.swap((None, Tag::None), Release) {
             barrier.reclaim(change);
         }
     }
 
+    /// Waits for the lock on the [`InternalNode`] to be released.
+    pub(super) fn wait(&self, barrier: &Barrier) {
+        let _result = self.wait_queue.wait(|| {
+            let ptr = self.latch.load(Relaxed, barrier);
+            if !ptr.is_null() || ptr.tag() == LOCKED {
+                // The `InternalNode` is being split or locked.
+                return Err(());
+            }
+            Ok(())
+        });
+    }
+
     /// Tries to coalesce nodes.
     fn coalesce<Q>(&self, removed_key: &Q, barrier: &Barrier) -> RemoveResult
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
     {
-        while let Some(lock) = Locker::try_lock(self) {
+        while let Some(lock) = Locker::try_lock(self, barrier) {
             let mut max_key_entry = None;
             for (key, node) in Scanner::new(&self.children) {
                 let node_ptr = node.load(Relaxed, barrier);
@@ -617,7 +639,7 @@ where
 
                     // Once the key is removed, it is safe to deallocate the node as the validation
                     // loop ensures the absence of readers.
-                    if let Some(node) = node.swap((None, Tag::None), Release) {
+                    if let Some(node) = node.swap((None, Tag::None), Release).0 {
                         barrier.reclaim(node);
                     }
                 } else {
@@ -630,22 +652,26 @@ where
             let fully_empty = if let Some(unbounded) = unbounded_ptr.as_ref() {
                 if unbounded.retired(Relaxed) {
                     if let Some((key, max_key_child)) = max_key_entry {
-                        if let Some(obsolete_node) = self.unbounded_child.swap(
-                            (max_key_child.get_arc(Relaxed, barrier), Tag::None),
-                            Release,
-                        ) {
+                        if let Some(obsolete_node) = self
+                            .unbounded_child
+                            .swap(
+                                (max_key_child.get_arc(Relaxed, barrier), Tag::None),
+                                Release,
+                            )
+                            .0
+                        {
                             debug_assert!(obsolete_node.retired(Relaxed));
                             barrier.reclaim(obsolete_node);
                         }
                         let result = self.children.remove_if(key.borrow(), &mut |_| true);
                         debug_assert_ne!(result, RemoveResult::Fail);
-                        if let Some(node) = max_key_child.swap((None, Tag::None), Release) {
+                        if let Some(node) = max_key_child.swap((None, Tag::None), Release).0 {
                             barrier.reclaim(node);
                         }
                         false
                     } else {
                         if let Some(obsolete_node) =
-                            self.unbounded_child.swap((None, RETIRED), Release)
+                            self.unbounded_child.swap((None, RETIRED), Release).0
                         {
                             debug_assert!(obsolete_node.retired(Relaxed));
                             barrier.reclaim(obsolete_node);
@@ -706,7 +732,7 @@ where
     K: 'static + Clone + Ord + Send + Sync,
     V: 'static + Clone + Send + Sync,
 {
-    lock: &'n AtomicArc<StructuralChange<K, V>>,
+    internal_node: &'n InternalNode<K, V>,
 }
 
 impl<'n, K, V> Locker<'n, K, V>
@@ -715,15 +741,16 @@ where
     V: Clone + Send + Sync,
 {
     /// Acquires exclusive lock on the [`InternalNode`].
-    pub fn try_lock(internal_node: &'n InternalNode<K, V>) -> Option<Locker<'n, K, V>> {
+    pub fn try_lock(
+        internal_node: &'n InternalNode<K, V>,
+        barrier: &'n Barrier,
+    ) -> Option<Locker<'n, K, V>> {
         if internal_node
             .latch
-            .compare_exchange(Ptr::null(), (None, LOCKED), Acquire, Relaxed)
+            .compare_exchange(Ptr::null(), (None, LOCKED), Acquire, Relaxed, barrier)
             .is_ok()
         {
-            Some(Locker {
-                lock: &internal_node.latch,
-            })
+            Some(Locker { internal_node })
         } else {
             None
         }
@@ -736,8 +763,9 @@ where
     V: Clone + Send + Sync,
 {
     fn drop(&mut self) {
-        debug_assert_eq!(self.lock.tag(Relaxed), LOCKED);
-        self.lock.swap((None, Tag::None), Release);
+        debug_assert_eq!(self.internal_node.latch.tag(Relaxed), LOCKED);
+        self.internal_node.latch.swap((None, Tag::None), Release);
+        self.internal_node.wait_queue.signal();
     }
 }
 
@@ -771,9 +799,11 @@ mod test {
                     children: Leaf::new(),
                     unbounded_child: AtomicArc::new(Node::new_leaf_node()),
                     latch: AtomicArc::null(),
+                    wait_queue: WaitQueue::default(),
                 }),
             }),
             latch: AtomicArc::null(),
+            wait_queue: WaitQueue::default(),
         }
     }
 
@@ -931,11 +961,15 @@ mod test {
                         {
                             barrier_clone.wait().await;
                             let barrier = Barrier::new();
-                            if let Ok(InsertResult::Success) =
-                                internal_node_clone.insert(fixed_point, fixed_point, &barrier)
-                            {
-                                assert!(!inserted_clone.swap(true, Relaxed));
-                            }
+                            match internal_node_clone.insert(fixed_point, fixed_point, &barrier) {
+                                Ok(InsertResult::Success) => {
+                                    assert!(!inserted_clone.swap(true, Relaxed));
+                                }
+                                Ok(InsertResult::Full(_, _) | InsertResult::Retired(_, _)) => {
+                                    internal_node_clone.rollback(&barrier);
+                                }
+                                _ => (),
+                            };
                             assert_eq!(
                                 internal_node_clone.search(&fixed_point, &barrier).unwrap(),
                                 &fixed_point
@@ -946,7 +980,12 @@ mod test {
                             let barrier = Barrier::new();
                             for i in 0..workload_size {
                                 if i != fixed_point {
-                                    let _result = internal_node_clone.insert(i, i, &barrier);
+                                    if let Ok(
+                                        InsertResult::Full(_, _) | InsertResult::Retired(_, _),
+                                    ) = internal_node_clone.insert(i, i, &barrier)
+                                    {
+                                        internal_node_clone.rollback(&barrier);
+                                    }
                                 }
                                 assert_eq!(
                                     internal_node_clone.search(&fixed_point, &barrier).unwrap(),
