@@ -2,6 +2,7 @@ use super::leaf::{InsertResult, RemoveResult, Scanner, DIMENSION};
 use super::Leaf;
 
 use crate::ebr::{Arc, AtomicArc, Barrier, Ptr, Tag};
+use crate::wait_queue::WaitQueue;
 use crate::LinkedList;
 
 use std::borrow::Borrow;
@@ -35,6 +36,9 @@ where
     /// `latch` acts as a mutex of the [`LeafNode`] that also stores the information about an
     /// on-going structural change.
     latch: AtomicArc<StructuralChange<K, V>>,
+
+    /// `wait_queue` for `latch`.
+    wait_queue: WaitQueue,
 }
 
 impl<K, V> LeafNode<K, V>
@@ -48,6 +52,7 @@ where
             children: Leaf::new(),
             unbounded_child: AtomicArc::null(),
             latch: AtomicArc::null(),
+            wait_queue: WaitQueue::default(),
         }
     }
 
@@ -206,6 +211,7 @@ where
                             }
                             InsertResult::Frozen(key, value) => {
                                 // The `Leaf` is being split: retry.
+                                self.wait(true, barrier);
                                 return Err((key, value));
                             }
                         };
@@ -222,6 +228,7 @@ where
                     (Some(Arc::new(Leaf::new())), Tag::None),
                     AcqRel,
                     Acquire,
+                    barrier,
                 ) {
                     Ok((_, ptr)) => {
                         unbounded_ptr = ptr;
@@ -252,6 +259,7 @@ where
                         );
                     }
                     InsertResult::Frozen(key, value) => {
+                        self.wait(true, barrier);
                         return Err((key, value));
                     }
                 };
@@ -286,6 +294,7 @@ where
                         if child.frozen() {
                             // When a `Leaf` is frozen, its entries may be being copied to new
                             // `Leaves`.
+                            self.wait(true, barrier);
                             return Err(result != RemoveResult::Fail);
                         }
                         if result == RemoveResult::Retired {
@@ -306,6 +315,7 @@ where
                 }
                 let result = unbounded.remove_if(key, condition);
                 if unbounded.frozen() {
+                    self.wait(true, barrier);
                     return Err(result != RemoveResult::Fail);
                 }
                 if result == RemoveResult::Retired {
@@ -322,6 +332,7 @@ where
     /// # Errors
     ///
     /// Returns an error if retry is required.
+    #[allow(clippy::too_many_lines)]
     fn split_leaf(
         &self,
         key: K,
@@ -331,7 +342,7 @@ where
         full_leaf: &AtomicArc<Leaf<K, V>>,
         barrier: &Barrier,
     ) -> Result<InsertResult<K, V>, (K, V)> {
-        let new_leaves = match self.latch.compare_exchange(
+        let new_leaves = if let Ok((_, ptr)) = self.latch.compare_exchange(
             Ptr::null(),
             (
                 Some(Arc::new(StructuralChange {
@@ -344,20 +355,27 @@ where
             ),
             Acquire,
             Relaxed,
+            barrier,
         ) {
-            Ok((_, ptr)) => {
-                if self.retired(Relaxed) {
-                    drop(self.latch.swap((None, Tag::None), Relaxed));
-                    return Ok(InsertResult::Retired(key, value));
-                }
-                if full_leaf_ptr != full_leaf.load(Relaxed, barrier) {
-                    drop(self.latch.swap((None, Tag::None), Relaxed));
-                    return Err((key, value));
-                }
-                ptr.as_ref().unwrap()
-            }
-            Err(_) => return Err((key, value)),
+            ptr.as_ref().unwrap()
+        } else {
+            self.wait(false, barrier);
+            return Err((key, value));
         };
+
+        if self.retired(Relaxed) {
+            let (change, _) = self.latch.swap((None, Tag::None), Relaxed);
+            self.wait_queue.signal();
+            drop(change);
+            return Ok(InsertResult::Retired(key, value));
+        }
+        if full_leaf_ptr != full_leaf.load(Relaxed, barrier) {
+            let (change, _) = self.latch.swap((None, Tag::None), Relaxed);
+            self.wait_queue.signal();
+            drop(change);
+            return Err((key, value));
+        }
+
         if let Some(full_leaf_key) = full_leaf_key {
             let ptr = addr_of!(new_leaves.origin_leaf_key) as *mut Option<K>;
             unsafe {
@@ -402,13 +420,12 @@ where
             let result =
                 target.push_back(low_key_leaf_ptr.get_arc().unwrap(), true, Release, barrier);
             debug_assert!(result.is_ok());
-            full_leaf.swap(
-                (
-                    new_leaves.low_key_leaf.swap((None, Tag::None), Relaxed),
-                    Tag::None,
-                ),
-                Release,
-            )
+            full_leaf
+                .swap(
+                    new_leaves.low_key_leaf.swap((None, Tag::None), Release),
+                    Release,
+                )
+                .0
         } else {
             // From here, `Scanners` can reach the new leaves.
             let result =
@@ -432,21 +449,26 @@ where
             };
 
             // Replace the full leaf with the high-key leaf.
-            let high_key_leaf = new_leaves.high_key_leaf.swap((None, Tag::None), Relaxed);
-            full_leaf.swap((high_key_leaf, Tag::None), Release)
+            let high_key_leaf = new_leaves.high_key_leaf.swap((None, Tag::None), Relaxed).0;
+            full_leaf.swap((high_key_leaf, Tag::None), Release).0
         };
 
+        // Unlocks the leaf node.
+        let (change, _) = self.latch.swap((None, Tag::None), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
+            barrier.reclaim(change);
+        }
+
         if let Some(unused_leaf) = unused_leaf {
+            // There is a possibility that entries were removed from the replaced leaf node whereas
+            // those entries still remain in `unused_leaf`; if that happens, iterators may see
+            // the removed entries momentarily.
             let deleted = unused_leaf.delete_self(Release);
             debug_assert!(deleted);
             barrier.reclaim(unused_leaf);
         }
 
-        // Unlocks the leaf node.
-        if let Some(change) = self.latch.swap((None, Tag::None), Release) {
-            barrier.reclaim(change);
-        }
-
         // Traverse several leaf nodes in order to cleanup deprecated links.
         self.max_le_appr(key.borrow(), barrier);
 
@@ -568,9 +590,11 @@ where
 
     /// Commits an on-going structural change.
     pub fn commit(&self, barrier: &Barrier) {
-        // Keeps the leaf node locked to prevent further locking attempts.
-        if let Some(change) = self.latch.swap((None, LOCKED), Relaxed) {
-            if let Some(obsolete_leaf) = change.origin_leaf.swap((None, Tag::None), Relaxed) {
+        // Mark the leaf node retired to prevent further locking attempts.
+        let (change, _) = self.latch.swap((None, RETIRED), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
+            if let Some(obsolete_leaf) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
                 // Makes the leaf unreachable before dropping it.
                 obsolete_leaf.delete_self(Relaxed);
                 barrier.reclaim(obsolete_leaf);
@@ -578,7 +602,7 @@ where
         }
     }
 
-    /// Rolls back the ongoing split operation recursively.
+    /// Rolls back the ongoing split operation.
     pub fn rollback(&self, barrier: &Barrier) {
         if let Some(change) = self.latch.load(Relaxed, barrier).as_ref() {
             let low_key_leaf_ptr = change.low_key_leaf.load(Relaxed, barrier);
@@ -596,12 +620,12 @@ where
                 debug_assert!(deleted);
             }
 
-            if let Some(origin) = change.origin_leaf.swap((None, Tag::None), Relaxed) {
+            if let Some(origin) = change.origin_leaf.swap((None, Tag::None), Relaxed).0 {
                 // Thaw the [`Leaf`].
                 let result = origin.thaw();
                 debug_assert!(result);
 
-                // Remove marks from the full leaf node.
+                // Remove the mark from the full leaf node.
                 //
                 // This unmarking has to be a release-store, otherwise it can be re-ordered
                 // before previous `delete_self` calls.
@@ -611,18 +635,37 @@ where
         }
 
         // Unlocks the leaf node.
-        if let Some(change) = self.latch.swap((None, Tag::None), Release) {
+        let (change, _) = self.latch.swap((None, Tag::None), Release);
+        self.wait_queue.signal();
+        if let Some(change) = change {
             barrier.reclaim(change);
         }
     }
 
+    /// Waits for the lock on the [`LeafNode`] to be released.
+    pub(super) fn wait(&self, expect_null: bool, barrier: &Barrier) {
+        let _result = self.wait_queue.wait(|| {
+            let ptr = self.latch.load(Relaxed, barrier);
+            let is_null = ptr.is_null();
+            if expect_null && is_null {
+                // The `LeafNode` may be locked, but it does not care.
+                return Ok(());
+            }
+            if !is_null || ptr.tag() == LOCKED {
+                // The `LeafNode` is being split or locked.
+                return Err(());
+            }
+            Ok(())
+        });
+    }
+
     /// Tries to coalesce empty or obsolete leaves after a successful removal of an entry.
     fn coalesce<Q>(&self, removed_key: &Q, barrier: &Barrier) -> RemoveResult
     where
         K: Borrow<Q>,
         Q: Ord + ?Sized,
     {
-        while let Some(lock) = Locker::try_lock(self) {
+        while let Some(lock) = Locker::try_lock(self, barrier) {
             let mut has_valid_leaf = false;
             for entry in Scanner::new(&self.children) {
                 let leaf_ptr = entry.1.load(Relaxed, barrier);
@@ -635,7 +678,7 @@ where
 
                     // The pointer is nullified after the metadata of `self.children` is updated so
                     // that readers are able to retry when they find it being `null`.
-                    if let Some(leaf) = entry.1.swap((None, Tag::None), Release) {
+                    if let Some(leaf) = entry.1.swap((None, Tag::None), Release).0 {
                         barrier.reclaim(leaf);
                     }
                 } else {
@@ -655,7 +698,7 @@ where
                         // It has to mark the pointer in order to prevent `LeafNode::insert` from
                         // replacing it with a new `Leaf`.
                         if let Some(obsolete_leaf) =
-                            self.unbounded_child.swap((None, RETIRED), Release)
+                            self.unbounded_child.swap((None, RETIRED), Release).0
                         {
                             barrier.reclaim(obsolete_leaf);
                         }
@@ -715,7 +758,7 @@ where
     K: 'static + Clone + Ord + Send + Sync,
     V: 'static + Clone + Send + Sync,
 {
-    lock: &'n AtomicArc<StructuralChange<K, V>>,
+    leaf_node: &'n LeafNode<K, V>,
 }
 
 impl<'n, K, V> Locker<'n, K, V>
@@ -724,15 +767,16 @@ where
     V: Clone + Send + Sync,
 {
     /// Acquires exclusive lock on the [`LeafNode`].
-    pub fn try_lock(leaf_node: &'n LeafNode<K, V>) -> Option<Locker<'n, K, V>> {
+    pub fn try_lock(
+        leaf_node: &'n LeafNode<K, V>,
+        barrier: &'n Barrier,
+    ) -> Option<Locker<'n, K, V>> {
         if leaf_node
             .latch
-            .compare_exchange(Ptr::null(), (None, LOCKED), Acquire, Relaxed)
+            .compare_exchange(Ptr::null(), (None, LOCKED), Acquire, Relaxed, barrier)
             .is_ok()
         {
-            Some(Locker {
-                lock: &leaf_node.latch,
-            })
+            Some(Locker { leaf_node })
         } else {
             None
         }
@@ -745,8 +789,9 @@ where
     V: Clone + Send + Sync,
 {
     fn drop(&mut self) {
-        debug_assert_eq!(self.lock.tag(Relaxed), LOCKED);
-        self.lock.swap((None, Tag::None), Release);
+        debug_assert_eq!(self.leaf_node.latch.tag(Relaxed), LOCKED);
+        self.leaf_node.latch.swap((None, Tag::None), Release);
+        self.leaf_node.wait_queue.signal();
     }
 }
 
@@ -964,18 +1009,27 @@ mod test {
                         {
                             barrier_clone.wait().await;
                             let barrier = Barrier::new();
-                            if let Ok(InsertResult::Success) =
-                                leaf_node_clone.insert(k, k, &barrier)
-                            {
-                                assert!(!inserted_clone.swap(true, Relaxed));
-                            }
+                            match leaf_node_clone.insert(k, k, &barrier) {
+                                Ok(InsertResult::Success) => {
+                                    assert!(!inserted_clone.swap(true, Relaxed));
+                                }
+                                Ok(InsertResult::Full(_, _) | InsertResult::Retired(_, _)) => {
+                                    leaf_node_clone.rollback(&barrier);
+                                }
+                                _ => (),
+                            };
                         }
                         {
                             barrier_clone.wait().await;
                             let barrier = Barrier::new();
                             for i in 0..workload_size {
                                 if i != k {
-                                    let _result = leaf_node_clone.insert(i, i, &barrier);
+                                    if let Ok(
+                                        InsertResult::Full(_, _) | InsertResult::Retired(_, _),
+                                    ) = leaf_node_clone.insert(i, i, &barrier)
+                                    {
+                                        leaf_node_clone.rollback(&barrier);
+                                    }
                                 }
                                 assert_eq!(leaf_node_clone.search(&k, &barrier).unwrap(), &k);
                             }
